import json
import time
from threading import Thread, Lock
import logging
from logging import Logger

from websocket import WebSocketApp

# TODO: Replace print statements with logging.


class WebsocketManager:
    _CONNECT_TIMEOUT_S = 5
    _ENDPOINT = None

    def __init__(self):
        self.connect_lock = Lock()
        self.ws = None
        self._logger: Logger = logging.getLogger('root')  # TODO: Integrate this with setup_custom_logger

    def _get_url(self):
        raise NotImplementedError()

    def _on_message(self, ws, message):
        raise NotImplementedError()

    def send(self, message):
        self.connect()
        self.ws.send(message)

    def send_json(self, message):
        self.send(json.dumps(message))

    def _connect(self):
        assert not self.ws, "ws should be closed before attempting to connect"

        self.ws = WebSocketApp(
            self._get_url(),
            on_message=self._wrap_callback(self._on_message),
            on_close=self._wrap_callback(self._on_close),
            on_error=self._wrap_callback(self._on_error),
        )

        wst = Thread(target=self._run_websocket, args=(self.ws,))
        wst.daemon = True
        wst.start()

        # Wait for socket to connect
        ts = time.time()
        while self.ws and (not self.ws.sock or not self.ws.sock.connected):
            if time.time() - ts > self._CONNECT_TIMEOUT_S:
                self.ws = None
                return

    def _wrap_callback(self, f):
        def wrapped_f(ws, *args, **kwargs):
            if ws is self.ws:
                try:
                    f(ws, *args, **kwargs)
                except Exception as e:
                    raise Exception(f'Error running websocket callback: {e}')
        return wrapped_f

    def _run_websocket(self, ws: WebSocketApp) -> None:
        try:
            ws.run_forever(ping_interval=15)
        except Exception as e:
            # raise Exception(f'Unexpected error while running websocket: {e}')
            self._logger.error(f'Unexpected error while running websocket: {e}')
            self._reconnect()
        finally:
            self._logger.debug("Error occurred. Function will return.")

    def _reconnect(self, ws: WebSocketApp) -> None:
        assert ws is not None, '_reconnect should only be called with an existing ws'
        if ws is self.ws:
            self.ws = None
            ws.close()
            self.connect()

    def close(self):
        if self.ws:
            self.ws.close()

    def connect(self) -> None:
        if self.ws:
            return
        with self.connect_lock:
            while not self.ws:
                self._connect()
                if self.ws:
                    self._logger.info(f"Connection with URL: {self._ENDPOINT}")
                    return

    def _on_close(self, ws: WebSocketApp, code, raw_msg) -> None:
        self._logger.debug(f"LedgerX: WebSocket connection closed: Code: {code}, Msg: {raw_msg}")
        if code:
            self._logger.debug("LedgerX: WebSocket reconnecting...")
            self._reconnect(ws)  # Only reconnect if there is an error code.

    def _on_error(self, ws: WebSocketApp, error) -> None:  # TODO: Add type hint for 'error'...
        self._logger.debug(f"LedgerX: WebSocket Error: {error}")
        self._reconnect(ws)

    def reconnect(self) -> None:
        if self.ws is not None:
            self._reconnect(self.ws)



import json
from collections import defaultdict
from typing import DefaultDict, List, Dict, Set
import time

from websocket_lx.websocket_manager import WebsocketManager


class LxWebsocketClient(WebsocketManager):
    _ENDPOINT = 'wss://api.ledgerx.com/ws'

    def __init__(self, api_key: str = None) -> None:
        super().__init__()
        self._api_key = api_key
        self._initial_open_order_count = 0
        self._last_heartbeat_timestamp: int = 0  # Nanosecond timestamp
        self._subscriptions: Set[int] = set()  # Do not reset orderbook top subscriptions
        self._fills: DefaultDict[str, int] = defaultdict(int)  # {message_id: filled_size}
        self._reset_data()

    def _on_open(self, ws):
        self._reset_data()

    def _reset_data(self) -> None:
        self._orderbook_timestamps: DefaultDict[int, float] = defaultdict(float)
        self._book_tops: DefaultDict[int, List[float]] = defaultdict(lambda: [0, 0, 0, 0])
        self._open_positions: DefaultDict[int, Dict[str, int]] = defaultdict(dict)
        self._account_balances: DefaultDict[str, Dict[str, float]] = defaultdict(dict)

        self._orderbook_timestamps.clear()
        self._book_tops.clear()
        self._open_positions.clear()
        self._account_balances.clear()

    def _get_url(self) -> str:
        if self._api_key:
            return self._ENDPOINT + f"?token={self._api_key}"
        else:
            return self._ENDPOINT

    def subscribe(self, contract_id: int) -> None:
        """Subscribe to orderbook tops for a given contract."""
        self._logger.info(f"LedgerX: Subscribed to orderbook top for contract_id: {contract_id}")
        self._subscriptions.add(contract_id)

    def subscribe_fills(self, message_id: str) -> None:
        """Subscribe to order fills for a given message_id."""
        self._logger.info(f"LedgerX: Subscribed to order fills for message_id: {message_id}")
        self._fills[message_id] = 0

    def unsubscribe(self, contract_id: int) -> None:
        """Unsubscribe from orderbook tops for a given contract."""
        self._logger.info(f"LedgerX: Unsubscribed from orderbook tops for contract_id: {contract_id}")
        self._subscriptions.remove(contract_id)

    def get_last_heartbeat_timestamp(self):
        return self._last_heartbeat_timestamp

    def get_book_top(self, contract_id: int) -> List[float]:
        """Returns the best [bid, bid_size, ask, ask_size] for a given contract_id."""
        return self._book_tops[contract_id]

    def get_book_timestamp(self, contract_id: int) -> float:
        """Returns local-time timestamp of last orderbook update for a given contract."""
        return self._orderbook_timestamps[contract_id]

    def get_open_position(self, contract_id: int) -> Dict:
        """Returns open position info for a given contract_id.
        https://docs.ledgerx.com/reference/market-data-feed#open_positions_update
        """
        return self._open_positions[contract_id]

    def get_filled_size(self, message_id: str) -> int:
        """Gets filled size for a given message_id. Make sure to subscribe_fills first.
        https://docs.ledgerx.com/reference/market-data-feed
        """
        return self._fills[message_id]

    def get_account_balances(self) -> Dict:
        """Returns account balances.
        https://docs.ledgerx.com/reference/market-data-feed#collateral_balance_update
        """
        return self._account_balances

    def _handle_state_manifest_message(self, message: Dict) -> None:
        self._initial_open_order_count = message['state_manifest']['open_order_count']
        self._logger.debug(f"LedgerX: Open Orders: {self._initial_open_order_count}")

    def _handle_open_positions_update_message(self, message: Dict) -> None:
        for position in message['positions']:
            contract_id = position['contract_id']
            del position['contract_id']
            self._open_positions.update({contract_id: position})

    def _handle_collateral_balance_message(self, message: Dict) -> None:
        """Updates account collateral balances."""
        def convert_units(symbol, value):
            """Converts from cents, satoshis, and gweis to dollars, BTC, and ETH respectively."""
            if symbol == 'USD':
                return value / 100
            elif symbol in ['BTC', 'CBTC']:
                return value / 10 ** 8
            elif symbol == 'ETH':
                return value / 10 ** 9

        # Convert currency units to USD, BTC, and ETH and update account balances.
        balances = message['collateral']
        for key in balances.keys():
            self._account_balances.update(
                {
                    key: {symbol: convert_units(symbol, value) for symbol, value in balances[key].items()}
                }
            )

        # TODO: Log account balances changed.

    def _handle_book_top_message(self, message: Dict) -> None:
        contract_id = message['contract_id']
        bid = message['bid']
        bid_size = message['bid_size']
        ask = message['ask']
        ask_size = message['ask_size']
        self._book_tops[contract_id] = [bid / 100, bid_size, ask / 100, ask_size]  # convert from cents to usd
        self._orderbook_timestamps[contract_id] = time.time()

    def _handle_heartbeat_message(self, message: Dict) -> None:
        """Records local-time of last received server heartbeat."""
        self._last_heartbeat_timestamp = time.time_ns()

    def _handle_fills_message(self, message: Dict) -> None:
        """Records amount filled for an order."""
        if 'filled_size' in message:
            self._fills[message['mid']] = message['filled_size']

    def _on_message(self, ws, raw_message: str) -> None:
        message = json.loads(raw_message)
        if 'type' not in message:
            return  # TODO: add handling for global state updates. track auth or unatuh status

        message_type = message['type']

        # Initial connection message. Includes open order count.
        if message_type == 'state_manifest':
            self._handle_state_manifest_message(message)

        # Account State Update: Updated list of account positions on all currently active contracts.
        if message_type == 'open_positions_update':
            self._handle_open_positions_update_message(message)

        # Account State Update: Current collateral balances.
        if message_type == 'collateral_balance_update':
            self._handle_collateral_balance_message(message)

        # Top of Book Feed: Top-of-orderbook for a given contract_id.
        if message_type == 'book_top':
            if message['contract_id'] in self._subscriptions:
                self._handle_book_top_message(message)

        # Exchange heartbeat. Monitors exchange life.
        if message_type == 'heartbeat':
            self._handle_heartbeat_message(message)

        if message_type == 'action_report':
            if message['mid'] in self._fills:
                self._handle_fills_message(message)



"""
LedgerX REST Client.
"""

from typing import Optional, Dict, Any, Tuple, List
from requests import Request, Session
from requests.exceptions import HTTPError, Timeout
import logging
from logging import Logger
import time
import json


class LxClient:
    _ENDPOINT = "https://api.ledgerx.com/"
    _TRADE_ENDPOINT = "https://trade.ledgerx.com/api/"

    def __init__(self, api_key=None) -> None:
        """Initializes LedgerX client with API key, http session, and logger."""
        self._session: Session = Session()
        self._api_key: str = api_key
        self._logger: Logger = logging.getLogger('root')  # TODO: Integrate this with setup_custom_logger
        self._retries = 0
        self._MAX_RETRIES = 3

    ############################################
    # HTTP Requests and Error Handling
    ############################################

    def _get(self, path: str, use_trade_api: bool = False, params: Optional[Dict[str, Any]] = None) -> Any:
        return self._request('GET', path, use_trade_api, params=params)

    def _post(self, path: str, use_trade_api: bool = False, params: Optional[Dict[str, Any]] = None) -> Any:
        return self._request('POST', path, use_trade_api, json=params)

    def _delete(self, path: str, use_trade_api: bool = False, params: Optional[Dict[str, Any]] = None) -> Any:
        return self._request('DELETE', path, use_trade_api, json=params)

    def _sign_request(self, request: Request) -> None:
        """Includes API Key in request headers."""
        if self._api_key:
            request.headers['Authorization'] = f'JWT {self._api_key}'

    def _request(self, method: str, path: str, use_trade_api: bool, **kwargs) -> Any:
        endpoint = self._ENDPOINT
        if use_trade_api:
            endpoint = self._TRADE_ENDPOINT
        url = endpoint + path
        request = Request(method, url, **kwargs)
        self._sign_request(request)
        prepared = request.prepare()
        response = self._session.send(prepared)

        # Request retry function.
        def retry():
            self._retries += 1
            if self._retries > self._MAX_RETRIES:
                raise Exception("Max retries on %s hit, raising. \n" % path +
                                "Request Parameters: % s \n" % json.dumps(kwargs) +
                                "Full url: %s" % url)
            else:
                return self._request(method, path, use_trade_api, **kwargs)

        # Response error-handling
        try:
            # Make non-200s throw
            response.raise_for_status()

        except HTTPError as e:
            # 400 - Bad Request.
            if response.status_code == 400:
                self._logger.error("400 bad request.")
                error = response.json()['error']

                if error == 'INVALID_TOKEN':
                    self._logger.error("API key incorrect or expired. Please check and restart.")
                    self._logger.error("Error: " + response.text)

                else:
                    self._logger.error("Error: " + response.text)

                exit(1)  # TODO: Replace all exit(1) calls w/ Exception classes

            # 401 - Unauthorized.
            if response.status_code == 401:
                self._logger.error("API Key unauthorized to perform the requested action.")
                self._logger.error("Error: " + response.text)
                exit(1)

            # 404 - Not Found.
            elif response.status_code == 404:
                self._logger.error("Unable to contact the LedgerX API (404)" +
                                   "Request: %s" % url)
                exit(1)

            # 429, Too Many Requests (ratelimit). Cancel orders & wait until rate limit reset.
            elif response.status_code == 429:
                self._logger.error("Ratelimited on current request. Sleeping, then trying again. " +
                                   "Try fewer order pairs. " +
                                   "Request: %s" % url)

                # Figure out how long we need to wait.
                sleep_time = int(response.headers['Retry-After'])  # seconds to wait.

                # Cancel orders.
                self._logger.warning("Canceling all orders.")
                self.cancel_all_orders()

                self._logger.error("Ratelimit will reset in %d seconds. Sleeping for %d seconds." % (sleep_time,
                                                                                                     sleep_time))
                time.sleep(sleep_time)

                # Retry the request
                return retry()

            # 503 - Service Unavailable. LedgerX temporary downtime.
            elif response.status_code == 503:
                self._logger.warning("Unable to contact the LedgerX API (503), retrying in 3 seconds. " +
                                     "Request: %s" % url)
                time.sleep(3)
                return retry()

            # If we haven't returned or re-raised yet, we get here.
            self._logger.error("Unhandled Error: %s: %s" % (e, response.text))
            self._logger.error("Endpoint was (%s)\n" % method +
                               "path was (%s):\n" % path +
                               "Request Parameters: %s" % json.dumps(kwargs))
            exit(1)

        except Timeout as e:
            # Timeout, re-run this request
            self._logger.warning("Timed out on request: %s" % url +
                                 "Keyword args: % s" % json.dumps(kwargs))
            return retry()

        # TODO: add ConnectionError retry logic

        # Reset retry counter on success
        self._retries = 0

        return response.json()

    ############################################
    # REST API Functions
    ############################################

    def list_contracts(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of contracts.
        https://docs.ledgerx.com/reference/listcontracts

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            contracts: dict
                List of contracts.
        """
        return self._get('trading/contracts', use_trade_api=False, params=params)

    def list_traded_contracts(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of contracts that you have traded.
        https://docs.ledgerx.com/reference/tradedcontracts

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            contracts: dict
                List of contracts.
        """
        return self._get('trading/contracts/traded', use_trade_api=False, params=params)

    def retrieve_contract(self, contract_id: int) -> dict:
        """Returns contract details for a single contract ID.
        https://docs.ledgerx.com/reference/retrievecontract

        Args:
            contract_id: int
                ID of record to fetch.

        Returns:
            contract: dict
                Contract details.
        """
        return self._get(f'trading/contracts/{contract_id}', use_trade_api=False)

    def retrieve_contract_position(self, contract_id: int) -> dict:
        """Returns your position for a given contract.
        https://docs.ledgerx.com/reference/positioncontract

        Args:
            contract_id: int
                ID of record to fetch.

        Returns:
            position: dict
                Position details.
        """
        return self._get(f'trading/contracts/{contract_id}/position', use_trade_api=False)

    def get_contract_ticker(self, contract_id: int, params: Optional[Dict[str, Any]] = None) -> dict:
        """Snapshot information about the current best bid/ask, 24h volume, and last trade.
        https://docs.ledgerx.com/reference/contractticker

        Args:
            contract_id: int
                ID of record to fetch.
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            ticker: dict
                Ticker information for the specified contract. All prices in cents.
        """
        return self._get(f'trading/contracts/{contract_id}/ticker', use_trade_api=False, params=params)

    def list_positions(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns all your positions.
        https://docs.ledgerx.com/reference/listpositions

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            contracts: dict
                List of positions.
        """
        return self._get('trading/positions', use_trade_api=False, params=params)

    def list_position_trades(self, contract_id: int, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of your trades for a given position.
        https://docs.ledgerx.com/reference/tradesposition

        Args:
            contract_id: int
                ID of record to fetch.
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            trades: dict
                List of trades.
        """
        return self._get(f'trading/positions/{contract_id}/trades', use_trade_api=False, params=params)

    def list_trades(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of your trades.
        https://docs.ledgerx.com/reference/listtrades

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            trades: dict
                List of trades.
        """
        return self._get('trading/trades', use_trade_api=False, params=params)

    def list_all_trades(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of all trades in the market.
        https://docs.ledgerx.com/reference/globalstrade

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            trades: dict
                List of trades.
        """
        return self._get('trading/trades/global', use_trade_api=False, params=params)

    def list_transactions(self, params: Optional[Dict[str, Any]] = None) -> dict:
        """Returns a list of all debits and credits to your accounts.
        https://docs.ledgerx.com/reference/gettransactions

        Args:
            params: dict
                Query parameters, {'param_name': value} pairings.

        Returns:
            transactions: dict
                List of transactions.
        """
        return self._get('funds/transactions', use_trade_api=False, params=params)

    def list_open_orders(self) -> dict:
        """Get all resting limit orders directly from the exchange.
        https://docs.ledgerx.com/reference/open-orders

        Returns:
            orders: dict
                Returns a list of your open orders.
        """
        return self._get('open-orders', use_trade_api=True)

    def create_order(self,
                     order_type: str,
                     contract_id: int,
                     is_ask: bool,
                     swap_purpose: str,
                     size: int,
                     price: int,
                     volatile: Optional[bool] = False) -> dict:
        """Place an order.
        https://docs.ledgerx.com/reference/create-order

        Returns:
            message_id: dict
                Returns the message id, or mid, of the order.
        """
        return self._post('orders', use_trade_api=True, params={
            'order_type': order_type,
            'contract_id': contract_id,
            'is_ask': is_ask,
            'swap_purpose': swap_purpose,
            'size': size,
            'price': price,
            'volatile': volatile
        })

    def cancel_all_orders(self) -> str:
        """Delete all outstanding orders associated with your MPID (the whole organization).
        https://docs.ledgerx.com/reference/cancel-all

        Returns:
            response_code: str
                200 Success, 400 Bad Request
        """
        return self._delete('orders', use_trade_api=True)

    def cancel_single_order(self, message_id: str, contract_id: int) -> str:
        """Cancel a single resting limit order.
        https://docs.ledgerx.com/reference/cancel-single

        Args:
            message_id: str
                The message id (mid) of the original order.
            contract_id: int
                The contract ID of the original order.

        Returns:
            response_code: str
                200 Success, 400 Bad Request
        """
        return self._delete(f'orders/{message_id}', use_trade_api=True, params={'contract_id': contract_id})

    def cancel_and_replace(self, message_id: str, contract_id: str, price: int, size: int) -> str:
        """Atomically swap an existing resting limit order with a new resting limit order. Price, side, and size may be changed.
        https://docs.ledgerx.com/reference/cancel-replace

        Args:
            message_id: str
                The message ID (mid) of the original order.
            contract_id: str
                The contract ID of the original order (cannot be changed in a cancel-replace).
            price: int
                The limit price of the new order in cents (USD) per contract. Must be a whole dollar amount.
            size: int
                How many units of the contract to transact for the new order.

        Returns:
            response_code: str
                200 Success, 400 Bad Request
        """
        return self._post(f'orders/{message_id}/edit', use_trade_api=True, params={
            'contract_id': contract_id,
            'price': price,
            'size': size
        })

    def get_contract_orderbook_state(self, contract_id: int):
        """Atomically swap an existing resting limit order with a new resting limit order. Price, side, and size may be changed.
        https://docs.ledgerx.com/reference/book-state-contract

        Args:
            contract_id: str
                The contract ID of the original order (cannot be changed in a cancel-replace).

        Returns:
            response_code: str
                200 Success, 400 Bad Request
        """
        return self._get(f'book-states/{contract_id}', use_trade_api=True)

    def get_account_balances(self) -> None:
        """Retrieves account balances. This is DEPRECATED, use WebSockets for account balance information instead."""
        return self._get('balance', use_trade_api=True)

    ############################################
    # Custom Functionality
    ############################################

    def get_swap_contract_info(self, asset: str) -> Dict:
        """Returns the contract info for the ETH or BTC next-day-swap contract."""
        assert asset in ['CBTC', 'ETH']
        contract_info = self.list_contracts({
            'active': True,
            'derivative_type': 'day_ahead_swap',
            'asset': asset
        })
        swap = contract_info['data'][0]
        return swap

    def get_swap_contract_id(self, asset: str) -> int:
        """Returns the contract_id for the ETH or BTC next-day-swap contract."""
        return self.get_swap_contract_info(asset)['id']

    def get_closest_call(self, asset: str) -> Tuple[int, int]:
        """Returns the contract_id and strike price for closest expiry ETH / BTC call option."""
        assert asset in ['CBTC', 'ETH']
        contract_info = self.list_contracts({
            'active': True,
            'contract_type': 'call',
            'derivative_type': 'options_contract',
            'asset': asset,
        })
        option = contract_info['data'][-1]
        return option['id'], option['strike_price'] / 100

    def get_book_top_rest(self, contract_id: int) -> List[int]:
        """Returns top of orderbook for a given contract_id. Uses REST endpoint."""
        states = self.get_contract_orderbook_state(contract_id)
        states = states['data']['book_states']
        asks = list(filter(lambda x: x['is_ask'], states))
        bids = list(filter(lambda x: not x['is_ask'], states))

        low_ask = asks[-1]
        top_bid = bids[0]

        bid = top_bid['price']
        bid_size = top_bid['size']

        ask = low_ask['price']
        ask_size = low_ask['size']
        return [bid, bid_size, ask, ask_size]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['rest_lx', 'websocket_lx']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.27.1,<3.0.0', 'websocket-client>=1.3.1,<2.0.0']

setup_kwargs = {
    'name': 'ftxusderivatives-python',
    'version': '0.1.2',
    'description': 'An unofficial Python wrapper for the REST and WebSocket APIs of FTX US Derivatives, formerly known as LedgerX.',
    'long_description': '# ftxusderivatives-python\nAn unofficial Python wrapper for the [REST and WebSocket APIs](https://docs.ledgerx.com/reference/overview) of FTX US Derivatives, formerly known as LedgerX. I have no affiliation with FTX US Derivatives. Use this at your own risk.\n\n## Features\n- Implementation of all REST endpoints, detailed [here](https://docs.ledgerx.com/reference/overview)\n- WebSocket implementation: live orderbook tops, account balances, open positions info, order fills, server heartbeat, reconnect logic\n- Simple handling of authentication\n- HTTP request error handling and retry logic\n- Logging support\n\n## Quick Start\n[Register an account with FTX US Derivatives.](https://derivs.ftx.us/) *[optional]*\n\n[Generate an API key](https://docs.ledgerx.com/docs/api-key) and configure permissions. *[optional]*\n\nInstall ftxusderviatives-python: `pip install ftxusderivatives-python`\n\nHere\'s some example code to get started with. Note that API calls that require authentication will not work if you do not\nenter your API key.\n\n```python\n###############################\n# REST API Example\n###############################\nfrom rest_lx.rest import LxClient\n\napi_key = ""  # TODO: Put API key here\n\n# Init REST client\nclient = LxClient(api_key=api_key)\n\n# list active day-ahead-swap contracts\nswaps = client.list_contracts({\n    \'active\': True,\n    \'derivative_type\': \'day_ahead_swap\',\n})\n\n# grab BTC day-ahead-swap contract ID\ndata = swaps[\'data\']\ncbtc_swap = filter(lambda data: data[\'underlying_asset\'] == \'CBTC\', data)\ncontract_id = next(cbtc_swap)[\'id\']\nprint(f"BTC swap contract_id: {contract_id}")\n\n# retrieve your position for BTC day-ahead-swap contract (requires authentication)\nposition = client.retrieve_contract_position(contract_id)\nprint(f"BTC swap position: {position}")\n\n# place bid for BTC next-day swap\nlx_buy = {\n    \'order_type\': \'limit\',\n    \'contract_id\': contract_id,\n    \'is_ask\': False,\n    \'swap_purpose\': \'undisclosed\',\n    \'size\': 1,\n    \'price\': 100,  # $1 (100 cents)\n    \'volatile\': True\n}\norder = client.create_order(**lx_buy)\n\n# cancel placed order\nmessage_id = order[\'data\'][\'mid\']  # order ID\nclient.cancel_single_order(message_id=message_id, contract_id=contract_id)\n\n###############################\n# WebSocket Example\n###############################\nfrom websocket_lx.client import LxWebsocketClient\nimport time\n\n# Init WebSocket client\nws = LxWebsocketClient(api_key=api_key)\n\n# Subscribe to orderbook-top feed for BTC day-ahead-swap contract\nws.subscribe(contract_id=contract_id)\nws.connect()\n\n# Grab orderbook-top for BTC day-ahead-swap once a second\nwhile True:\n    top = ws.get_book_top(contract_id=contract_id)\n    print(top)\n    time.sleep(1)\n```\n\n## Todo\n- Repo documentation\n- [Order fills, cancels, and insertions support](https://docs.ledgerx.com/reference/market-data-feed)\n\n## Contributing \nContributions, fixes, recommendations, and all other feedback is welcome. If you are fixing a bug, please open an issue first with all relevant details, and mention the issue number in the pull request.\n\n### Contact \nI can be reached on discord at Nenye#5335, or through email at nenye@ndili.net. Otherwise, feel free to open a PR or Issue here.\n',
    'author': 'Nenye Ndili',
    'author_email': 'nenye@ndili.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nenyehub/ftxusderivatives-python',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
